@@ -25,18 +25,32 @@
 
 #include "common.h"
 
+#ifdef HAVE_LIBZ
+#include <zlib.h>
+#endif
+
+#ifdef HAVE_LIBSSL
 typedef struct ssl_st SSL;
 typedef struct ssl_ctx_st SSL_CTX;
 typedef struct stack_st _STACK;
 
+enum {
+	SSLv2 = 1,
+	SSLv3 = 2,
+	TLSv1 = 4,
+	TLSv1_1 = 8,
+	TLSv1_2 = 16
+};
+#endif
+
 typedef struct server_conf {
 	char *tunnel;
 	char *host;
 	int port;
 #ifdef HAVE_LIBSSL
 	char *cert_file;
-	char use_imaps;
-	char use_sslv2, use_sslv3, use_tlsv1, use_tlsv11, use_tlsv12;
+	char system_certs;
+	char ssl_versions;
 
 	/* these are actually variables and are leaked at the end */
 	char ssl_ctx_valid;
@@ -47,9 +61,8 @@ typedef struct server_conf {
 
 typedef struct buff_chunk {
 	struct buff_chunk *next;
-	char *data;
 	int len;
-	char buf[1];
+	char data[1];
 } buff_chunk_t;
 
 typedef struct {
@@ -65,27 +78,44 @@ typedef struct {
 	char *name;
 #ifdef HAVE_LIBSSL
 	SSL *ssl;
-	int force_trusted;
+	wakeup_t ssl_fake;
+#endif
+#ifdef HAVE_LIBZ
+	z_streamp in_z, out_z;
+	wakeup_t z_fake;
+	int z_written;
 #endif
 
 	void (*bad_callback)( void *aux ); /* async fail while sending or listening */
 	void (*read_callback)( void *aux ); /* data available for reading */
-	int (*write_callback)( void *aux ); /* all *queued* data was sent */
+	void (*write_callback)( void *aux ); /* all *queued* data was sent */
 	union {
 		void (*connect)( int ok, void *aux );
 		void (*starttls)( int ok, void *aux );
 	} callbacks;
 	void *callback_aux;
 
+	notifier_t notify;
+	wakeup_t fd_fake;
+
 	/* writing */
+	buff_chunk_t *append_buf; /* accumulating buffer */
 	buff_chunk_t *write_buf, **write_buf_append; /* buffer head & tail */
+	int writing;
+#ifdef HAVE_LIBZ
+	int append_avail; /* space left in accumulating buffer */
+#endif
 	int write_offset; /* offset into buffer head */
+	int buffer_mem; /* memory currently occupied by buffers in the queue */
 
 	/* reading */
 	int offset; /* start of filled bytes in buffer */
 	int bytes; /* number of filled bytes in buffer */
 	int scanoff; /* offset to continue scanning for newline at, relative to 'offset' */
 	char buf[100000];
+#ifdef HAVE_LIBZ
+	char z_buf[100000];
+#endif
 } conn_t;
 
 /* call this before doing anything with the socket */
@@ -93,7 +123,7 @@ static INLINE void socket_init( conn_t *conn,
                                 const server_conf_t *conf,
                                 void (*bad_callback)( void *aux ),
                                 void (*read_callback)( void *aux ),
-                                int (*write_callback)( void *aux ),
+                                void (*write_callback)( void *aux ),
                                 void *aux )
 {
 	conn->conf = conf;
@@ -107,13 +137,16 @@ static INLINE void socket_init( conn_t *conn,
 }
 void socket_connect( conn_t *conn, void (*cb)( int ok, void *aux ) );
 void socket_start_tls(conn_t *conn, void (*cb)( int ok, void *aux ) );
+void socket_start_deflate( conn_t *conn );
 void socket_close( conn_t *sock );
 int socket_read( conn_t *sock, char *buf, int len ); /* never waits */
 char *socket_read_line( conn_t *sock ); /* don't free return value; never waits */
 typedef enum { KeepOwn = 0, GiveOwn } ownership_t;
-int socket_write( conn_t *sock, char *buf, int len, ownership_t takeOwn );
-
-void cram( const char *challenge, const char *user, const char *pass,
-           char **_final, int *_finallen );
+typedef struct conn_iovec {
+	char *buf;
+	int len;
+	ownership_t takeOwn;
+} conn_iovec_t;
+void socket_write( conn_t *sock, conn_iovec_t *iov, int iovcnt );
 
 #endif
